Unit GLZVectorMath;

{$mode objfpc}{$H+}
{.$DEFINE USE_ASM}
{.$DEFINE USE_SSE_ASM}
{.$DEFINE USE_AVX_ASM}
{$MODESWITCH ADVANCEDRECORDS}

{ Some informations
 From what I read,
   - On Windows some 32-bit compiler put the "Self" in ECX, result is put in EDX
   - On Windows some 64-bit compiler put the "Self" in RCX, result in most case is put in RDX ( or RBP-8, but this is not safe)
   - On Linux 32bits most compiler put the "Self" in ??? EDI ???
   - On Linux 64bits most compiler put the "Self" in RDI

     So on Linux we need mov it mov edx,esi / mov rcx, rdi
   - With AVX if we use YMM register we need call VZEROUPPER before send the result
   - For Single result for avoir compiler warning we use MOVLPS instead MOVUPS both with SSE and AVX
   - We loose a little precision by using Sqrt RSqrt SSE/AVX function
}

Interface

Uses
  Classes, Sysutils;

type
  TGLZVector3fType = packed array[0..2] of Single;
  TGLZVector4fType = packed array[0..3] of Single;

  TGLZVector3f = packed record
    case Byte of
      0: (X, Y, Z: Single);
      1: (V: TGLZVector3fType);
  End;

  TGLZSSEVector4f = packed record
    public
      procedure Create(Const aX,aY,aZ,aW : Single);
      function ToString : String;

      class operator +(constref A, B: TGLZSSEVector4f): TGLZSSEVector4f; overload;
      class operator -(constref A, B: TGLZSSEVector4f): TGLZSSEVector4f; overload;
      class operator *(constref A, B: TGLZSSEVector4f): TGLZSSEVector4f; overload;
      class operator /(constref A, B: TGLZSSEVector4f): TGLZSSEVector4f; overload;

      class operator +(constref A: TGLZSSEVector4f; constref B:Single): TGLZSSEVector4f; overload;
      class operator -(constref A: TGLZSSEVector4f; constref B:Single): TGLZSSEVector4f; overload;
      class operator *(constref A: TGLZSSEVector4f; constref B:Single): TGLZSSEVector4f; overload;
      class operator /(constref A: TGLZSSEVector4f; constref B:Single): TGLZSSEVector4f; overload;

      class operator -(constref A: TGLZSSEVector4f): TGLZSSEVector4f; overload;

      //function Negate :TGLZVector4f;
      function DivideBy2:TGLZSSEVector4f;

      function Distance(constref A: TGLZSSEVector4f):Single;
      function Length:Single;

      function CrossProduct(constref A: TGLZSSEVector4f): TGLZSSEVector4f;
      function DotProduct(constref A: TGLZSSEVector4f):Single;


      function Normalize: TGLZSSEVector4f;
      function Norm:Single;

      case Byte of
        0: (X, Y, Z, W: Single);
        1: (V: TGLZVector4fType);
        2: (AsVector3f : TGLZVector3f);
  end;

  TGLZAVXVector4f = packed record
    public
      procedure Create(Const aX,aY,aZ,aW : Single);
      function ToString : String;

      class operator +(constref A, B: TGLZAVXVector4f): TGLZAVXVector4f; overload;
      class operator -(constref A, B: TGLZAVXVector4f): TGLZAVXVector4f; overload;
      class operator *(constref A, B: TGLZAVXVector4f): TGLZAVXVector4f; overload;
      class operator /(constref A, B: TGLZAVXVector4f): TGLZAVXVector4f; overload;

      class operator +(constref A: TGLZAVXVector4f; constref B:Single): TGLZAVXVector4f; overload;
      class operator -(constref A: TGLZAVXVector4f; constref B:Single): TGLZAVXVector4f; overload;
      class operator *(constref A: TGLZAVXVector4f; constref B:Single): TGLZAVXVector4f; overload;
      class operator /(constref A: TGLZAVXVector4f; constref B:Single): TGLZAVXVector4f; overload;

      class operator -(constref A: TGLZAVXVector4f): TGLZAVXVector4f; overload;

      function DivideBy2:TGLZAVXVector4f;

      function Distance(constref A: TGLZAVXVector4f):Single;
      function Length:Single;

      function CrossProduct(constref A: TGLZAVXVector4f): TGLZAVXVector4f;
      function DotProduct(constref A: TGLZAVXVector4f):Single;
      function Norm:Single;
      function Normalize: TGLZAVXVector4f;


      case Byte of
        0: (X, Y, Z, W: Single);
        1: (V: TGLZVector4fType);
        2: (AsVector3f : TGLZVector3f);
  end;

  TGLZVector4f = packed record
    public
      procedure Create(Const aX,aY,aZ,aW : Single);
      function ToString : String;

      class operator +(constref A, B: TGLZVector4f): TGLZVector4f; overload;
      class operator -(constref A, B: TGLZVector4f): TGLZVector4f; overload;
      class operator *(constref A, B: TGLZVector4f): TGLZVector4f; overload;
      class operator /(constref A, B: TGLZVector4f): TGLZVector4f; overload;

      class operator +(constref A: TGLZVector4f; constref B:Single): TGLZVector4f; overload;
      class operator -(constref A: TGLZVector4f; constref B:Single): TGLZVector4f; overload;
      class operator *(constref A: TGLZVector4f; constref B:Single): TGLZVector4f; overload;
      class operator /(constref A: TGLZVector4f; constref B:Single): TGLZVector4f; overload;

      class operator -(constref A: TGLZVector4f): TGLZVector4f; overload;

      //function Negate :TGLZVector4f;

      function DivideBy2:TGLZVector4f;

      function Distance(constref A: TGLZVector4f):Single;
      function Length:Single;

      function DotProduct(constref A: TGLZVector4f):Single;
      function CrossProduct(constref A: TGLZVector4f): TGLZVector4f;

      function Normalize: TGLZVector4f;
      function Norm:Single;

      case Byte of
        0: (X, Y, Z, W: Single);
        1: (V: TGLZVector4fType);
        2: (AsVector3f : TGLZVector3f);
  end;

Const
  cNullVector4f    : TGLZVector4f    = (x:0;y:0;z:0;w:0);
  cNullSSEVector4f : TGLZSSEVector4f = (x:0;y:0;z:0;w:0);
  cNullAVXVector4f : TGLZAVXVector4f = (x:0;y:0;z:0;w:0);

  cOneVector4f    : TGLZVector4f    = (x:1;y:1;z:1;w:1);
  cOneSSEVector4f : TGLZSSEVector4f = (x:1;y:1;z:1;w:1);
  cOneAVXVector4f : TGLZAVXVector4f = (x:1;y:1;z:1;w:1);

  cSSE_MASK_NO_W   : TGLZSSEVector4f = (v:($FFFFFFFF,  $FFFFFFFF, $FFFFFFFF,$00000000 ));


Implementation

{.$IFDEF USE_ASM}
{%region%====[ AVX Vectors Functions ]==========================================}
{.$IFDEF USE_AVX_ASM}

procedure TGLZAVXVector4f.Create(Const aX,aY,aZ,aW : Single);
begin
   Self.X := AX;
   Self.Y := AY;
   Self.Z := AZ;
   Self.W := AW;
end;

function TGLZAVXVector4f.ToString : String;
begin
   Result := '(X: '+FloattoStrF(Self.X,fffixed,5,5)+
            ' ,Y: '+FloattoStrF(Self.Y,fffixed,5,5)+
            ' ,Z: '+FloattoStrF(Self.Z,fffixed,5,5)+
            ' ,W: '+FloattoStrF(Self.W,fffixed,5,5)+')';
End;

class operator TGLZAVXVector4f.+(constref A, B: TGLZAVXVector4f): TGLZAVXVector4f; assembler;
asm
  VMOVUPS XMM0,[A]
  VMOVUPS XMM1,[B]
  VADDPS  XMM0,XMM0, XMM1
  VMOVUPS [RESULT], XMM0
end;

class operator TGLZAVXVector4f.-(constref A, B: TGLZAVXVector4f): TGLZAVXVector4f; assembler;
asm
  VMOVUPS XMM0,[A]
  VMOVUPS XMM1,[B]
  VSUBPS  XMM0,XMM0, XMM1
  VMOVUPS [RESULT], XMM0
end;

class operator TGLZAVXVector4f.*(constref A, B: TGLZAVXVector4f): TGLZAVXVector4f; assembler;
asm
  VMOVUPS XMM0,[A]
  VMOVUPS XMM1,[B]
  VMULPS  XMM0,XMM0, XMM1
  VMOVUPS [RESULT], XMM0
end;

class operator TGLZAVXVector4f./(constref A, B: TGLZAVXVector4f): TGLZAVXVector4f; assembler;
asm
  VMOVUPS XMM0,[A]
  VMOVUPS XMM1,[B]
  VDIVPS  XMM0,XMM0, XMM1
  VMOVUPS [RESULT], XMM0
end;

class operator TGLZAVXVector4f.+(constref A: TGLZAVXVector4f; constref B:Single): TGLZAVXVector4f; assembler;
asm
  VMOVUPS XMM0,[A]
  VMOVSS  XMM1,[B]
  VSHUFPS XMM1, XMM1, XMM1,0
  VADDPS  XMM0,XMM1, XMM0
  VMOVUPS [RESULT], XMM0
end;

class operator TGLZAVXVector4f.-(constref A: TGLZAVXVector4f; constref B:Single): TGLZAVXVector4f; assembler;
asm
  VMOVUPS XMM0,[A]
  VMOVSS  XMM1,[B]
  VSHUFPS XMM1, XMM1, XMM1,0
  VSUBPS  XMM0,XMM0, XMM1
  VMOVUPS [RESULT], XMM0
end;

class operator TGLZAVXVector4f.*(constref A: TGLZAVXVector4f; constref B:Single): TGLZAVXVector4f; assembler;
asm
  VMOVUPS XMM0,[A]
  VMOVSS  XMM1,[B]
  VSHUFPS XMM1, XMM1, XMM1,0
  VMULPS  XMM0,XMM1, XMM0
  VMOVUPS [RESULT], XMM0
end;

class operator TGLZAVXVector4f./(constref A: TGLZAVXVector4f; constref B:Single): TGLZAVXVector4f; assembler;
asm
  VMOVUPS XMM0,[A]
  VMOVSS  XMM1,[B]
  VSHUFPS XMM1, XMM1, XMM1,0
  VDIVPS  XMM0,XMM0, XMM1
  VMOVUPS [RESULT], XMM0
end;

class operator TGLZAVXVector4f.-(constref A: TGLZAVXVector4f): TGLZAVXVector4f; assembler;
asm
  vmovups xmm1,[A]    // RCX = Self
  vmovups xmm0,[RIP+cNullAVXVector4f]
  VSUBPS  XMM0,XMM0, XMM1
  vmovups [Result],xmm0 //RDX = Result
End;

function TGLZAVXVector4f.DivideBy2:TGLZAVXVector4f;assembler;
const
  sub_value: TGLZAVXVector4f = (X:0.5; Y:0.5; Z:0.5; W:0.5); //  00800000h,00800000h,00800000h,00800000h
asm
  vmovups xmm0,[RCX]
  vmovups xmm1,[RIP+sub_value]
  vmulps xmm0, xmm0,xmm1
  vmovups [Result],xmm0
end;

function TGLZAVXVector4f.Distance(constref A: TGLZAVXVector4f):Single;assembler;
Asm
  vmovups  xmm0, [RCX]
  vmovups  xmm1, [A]
  vsubps   xmm0, xmm0, xmm1
  vmulps   xmm0, xmm0,xmm0
  vmovaps  xmm1,xmm0;
  vshufps  xmm1,xmm1,xmm1,11100001b
  vmovaps  xmm2,xmm0
  vshufps  xmm2,xmm2,xmm2,11000110b
  vaddss   xmm0,xmm0,xmm1
  vaddss   xmm0,xmm0,xmm2
  vsqrtss  xmm0,xmm0,xmm0
  MOVLPS [Result], XMM0
end;

function TGLZAVXVector4f.Length:Single;assembler;
Asm
  vmovups  xmm0, [RCX]
  vmulps   xmm0,xmm0, xmm0
  vmovaps  xmm1,xmm0
  vmovaps  xmm2,xmm0
  vshufps  xmm1, xmm1,xmm1,$E1
  vshufps  xmm2,xmm2, xmm1,$C6
  vaddps   xmm2, xmm2,xmm0
  vaddps   xmm2, xmm2,xmm1
  vSQRTSS  xmm0, XMM0, XMM2
  MOVLPS [Result], XMM0
end;

function TGLZAVXVector4f.CrossProduct(constref A: TGLZAVXVector4f): TGLZAVXVector4f;assembler;
asm
  VMOVUPS XMM0, [RCX]               // RCX = Self

  VMOVUPS XMM1, [A]
  VMOVAPS XMM2, XMM0
  VMOVAPS XMM3, XMM1


  VSHUFPS XMM0, XMM0, XMM0, $C9

  VSHUFPS XMM1, XMM1, XMM1, $D2
  VSHUFPS XMM2, XMM2, XMM2, $D2
  VSHUFPS XMM3, XMM3, XMM3, $C9

  VMULPS  XMM0, XMM0, XMM1
  VMULPS  XMM2, XMM2, XMM3

  VSUBPS  XMM0, XMM0, XMM2

  MOVUPS [RESULT], XMM0
end;

function TGLZAVXVector4f.DotProduct(constref A: TGLZAVXVector4f):Single;assembler;
asm
  vmovups xmm0, [RCX]
  vmovups xmm1, [A]
  vdpps xmm0, xmm0, xmm1, 01110001b //or $F1
  movlps [Result], xmm0
end;

function TGLZAVXVector4f.Norm:Single;assembler;
asm
  vmovaps xmm0, [RCX]

  vmulps xmm0,xmm0, xmm0
  vmovaps xmm1, xmm0
  vshufps xmm0, xmm0, xmm1, $4e
  vaddps xmm0,xmm0, xmm1

  vshufps xmm1, xmm1, xmm0, $11
  vaddps xmm0, xmm0, xmm1
  movlps [RESULT], xmm0
end;


function TGLZAVXVector4f.Normalize: TGLZAVXVector4f;  assembler;
asm
  vmovups xmm0, [RCX]
  vmovaps xmm2, xmm0
  vmulps xmm0, xmm0, xmm0
  vmovaps xmm1, xmm0
  vshufps xmm0, xmm0, xmm1, $4e
  vaddps xmm0,xmm0, xmm1
  vmovaps xmm1, xmm0
  vshufps xmm1, xmm1, xmm1, $11
  vaddps xmm0, xmm0, xmm1
  vrsqrtps xmm0, xmm0
  vmulps xmm2, xmm2, xmm0
  vmovups [Result], xmm2
end;



{.$ENDIF}
{%endregion%}
{.$IFDEF USE_SSE_ASM}
{%region%====[ SSE Vectors Functions ]==========================================}

procedure TGLZSSEVector4f.Create(Const aX,aY,aZ,aW : Single);
begin
   Self.X := AX;
   Self.Y := AY;
   Self.Z := AZ;
   Self.W := AW;
end;

function TGLZSSEVector4f.ToString : String;
begin
   Result := '(X: '+FloattoStrF(Self.X,fffixed,5,5)+
            ' ,Y: '+FloattoStrF(Self.Y,fffixed,5,5)+
            ' ,Z: '+FloattoStrF(Self.Z,fffixed,5,5)+
            ' ,W: '+FloattoStrF(Self.W,fffixed,5,5)+')';
End;

class operator TGLZSSEVector4f.+(constref A, B: TGLZSSEVector4f): TGLZSSEVector4f; assembler;
asm
  MOVUPS XMM0,[A]
  MOVUPS XMM1,[B]
  ADDPS  XMM0,XMM1
  MOVUPS [RESULT], XMM0
end;

class operator TGLZSSEVector4f.-(constref A, B: TGLZSSEVector4f): TGLZSSEVector4f; assembler;
asm
  MOVUPS XMM0,[A]
  MOVUPS XMM1,[B]
  SUBPS  XMM0,XMM1
  MOVUPS [RESULT], XMM0
end;

class operator TGLZSSEVector4f.*(constref A, B: TGLZSSEVector4f): TGLZSSEVector4f; assembler;
asm
  MOVUPS XMM0,[A]
  MOVUPS XMM1,[B]
  MULPS  XMM0,XMM1
  MOVUPS [RESULT], XMM0
end;

class operator TGLZSSEVector4f./(constref A, B: TGLZSSEVector4f): TGLZSSEVector4f; assembler;
asm
  MOVUPS XMM0,[A]
  MOVUPS XMM1,[B]
  DIVPS  XMM0,XMM1
  MOVUPS [RESULT], XMM0
end;

class operator TGLZSSEVector4f.+(constref A: TGLZSSEVector4f; constref B:Single): TGLZSSEVector4f; assembler;
asm
  MOVUPS XMM0,[A]
  MOVSS  XMM1,[B]
  SHUFPS XMM1, XMM1,0
  ADDPS  XMM0,XMM1
  MOVUPS [RESULT], XMM0
end;

class operator TGLZSSEVector4f.-(constref A: TGLZSSEVector4f; constref B:Single): TGLZSSEVector4f; assembler;
asm
  MOVUPS XMM0,[A]
  MOVSS  XMM1,[B]
  SHUFPS XMM1, XMM1,0
  SUBPS  XMM0,XMM1
  MOVUPS [RESULT], XMM0
end;

class operator TGLZSSEVector4f.*(constref A: TGLZSSEVector4f; constref B:Single): TGLZSSEVector4f; assembler;
asm
  MOVUPS XMM0,[A]
  MOVSS  XMM1,[B]
  SHUFPS XMM1, XMM1,0
  MULPS  XMM0,XMM1
  MOVUPS [RESULT], XMM0
end;

class operator TGLZSSEVector4f./(constref A: TGLZSSEVector4f; constref B:Single): TGLZSSEVector4f; assembler;
asm
  MOVUPS XMM0,[A]
  MOVSS  XMM1,[B]
  SHUFPS XMM1, XMM1,0
  DIVPS  XMM0,XMM1
  MOVUPS [RESULT], XMM0
end;

class operator TGLZSSEVector4f.-(constref A: TGLZSSEVector4f): TGLZSSEVector4f; assembler;
asm
    movups xmm1,[A]    // RCX = Self
    movups xmm0,[RIP+cNullSSEVector4f]
    subps xmm0,xmm1
    movups [Result],xmm0 //RBX = Result
End;

function TGLZSSEVector4f.DivideBy2:TGLZSSEVector4f;assembler;
const
  sub_value: TGLZSSEVector4f = (X:0.5; Y:0.5; Z:0.5; W:0.5); //  00800000h,00800000h,00800000h,00800000h
asm
  movups xmm0,[RCX]
  movups xmm1,[RIP+sub_value]
  mulps xmm0,xmm1
  movups [Result],xmm0
end;

function TGLZSSEVector4f.Distance(constref A: TGLZSSEVector4f):Single;assembler;
Asm
  movups  xmm0, [RCX] //RCX = SELF
  movups  xmm1, [A]
  subps   xmm0, xmm1
  mulps   xmm0,xmm0
  movaps  xmm1,xmm0;
  shufps  xmm1,xmm1,11100001b
  movaps  xmm2,xmm0
  shufps  xmm2,xmm2,11000110b
  addss   xmm0,xmm1
  addss   xmm0,xmm2
  sqrtss  xmm0,xmm0
  MOVLPS [Result], XMM0  //RBP-8 = RESULT
end;

function TGLZSSEVector4f.Length:Single;assembler;
Asm
  movups  xmm0, [RCX]
  mulps   xmm0, xmm0
  movaps  xmm1,xmm0
  movaps  xmm2,xmm0
  shufps  xmm1,xmm1,$E1
  shufps  xmm2,xmm1,$C6
  addps   xmm2,xmm0
  addps   xmm2,xmm1
  SQRTSS  XMM0, XMM2
  MOVLPS [Result], XMM0
end;

function TGLZSSEVector4f.CrossProduct(constref A: TGLZSSEVector4f): TGLZSSEVector4f;assembler;
asm
  // Place unaligned data to registers
 //v1:x,y,z,w
  MOVUPS XMM0, [RCX]               // RCX = Self
 //v2:x,y,z,w
  MOVUPS XMM1, [A]                // XMM1 = V2
  MOVAPS XMM2, XMM0                // XMM2 = V1 // move aligned data
  MOVAPS XMM3, XMM1                // XMM3 = V2

  // Shuffle members
  // Becarefull at the order in the register (right to left )= w z x y )
  // 1st arg XMM = w z y x = 11 10 01 00
  // 2nd arg XMM = w z y x = 11 10 01 00
  // 3rd Arg where to place members of 1st arg in 2nd Arg
  // Becarefull reading from Right to Left
  SHUFPS XMM0, XMM0, $C9  //w z y x -> $C9 = 11 00 10 01 -> w x z y

  SHUFPS XMM1, XMM1, $D2  //w z y x -> $D2 = 11 01 00 10 -> w y x z
  SHUFPS XMM2, XMM2, $D2
  SHUFPS XMM3, XMM3, $C9
  // Scale
  MULPS  XMM0, XMM1
  MULPS  XMM2, XMM3
  // Sub the results
  SUBPS  XMM0, XMM2

  MOVUPS [RESULT], XMM0      // return result
end;

// if the return is > 0 then the angle between the 2 vectors is  > 90° else is < 90°
// Usefull for conputing backface culling and lighting
function TGLZSSEVector4f.DotProduct(constref A: TGLZSSEVector4f):Single;assembler;
asm
  movups xmm0, [RCX]
  movups xmm1, [A]

  // SSE version 1 not optimal
(*  mulps xmm0, xmm1

  movaps  xmm1,xmm0
  movaps  xmm2,xmm0

  shufps  xmm1,xmm1,$E1
  shufps  xmm2,xmm1,$D2

  addps   xmm0,xmm1
  addps   xmm0,xmm2

  movlps [Result], xmm0 *)

  // SSE 2
  mulps  xmm0, xmm1             //   xmm0 =  w | z  | y  | x  |
  movhlps xmm1, xmm0            //   xmm1 =  - | -  | w  | z  |
  addps xmm1, xmm0              // x + z
  // How to with shufps
  //  In xmm0 =  w | z  | y  | x  |
  //          = 11 | 10 | 01 | 00 |
  // Out xmm0 =  - | -  | -  | Y  |
  //          = 00 | 00 | 00 | 01 | ==> 00000001b
  shufps xmm0, xmm0, 00000001b  //   xmm1 =  - | -  | -  | y  |
  addps xmm0, xmm1              // (x + z ) + y
  movlps [Result], xmm0


  //  SSE 3 version
 (* mulps xmm1, xmm0
  // Sum components
  haddps xmm1, xmm1
  haddps xmm1, xmm1

  movlps [Result], xmm1 *)

  // SSE 4
 (*
   // Becarefull at the order of mask Right to Left
   // 4 high bits: which elements should be summed. (w,z,y,x)
   // 4 low bits: which output slots should contain the result. (3,2,1,0)
   // mask =  0111b 0001;
   dpps xmm0, xmm1, 01110001b //or $F1
   movlps [Result], xmm0 *)

end;

function TGLZSSEVector4f.Norm:Single;assembler;
asm
  movaps xmm0, [RCX]
  mulps xmm0, xmm0
  movaps xmm1, xmm0
  shufps xmm0, xmm1, $4e
  addps xmm0, xmm1
  shufps xmm1, xmm0, $11
  addps xmm0, xmm1
  movlps [RESULT], xmm0
end;

function TGLZSSEVector4f.Normalize: TGLZSSEVector4f;  assembler;
asm
  movups xmm0, [RCX]
  movaps xmm2, xmm0
  mulps xmm0, xmm0
  movaps xmm1, xmm0
  shufps xmm0, xmm1, $4e
  addps xmm0, xmm1
  movaps xmm1, xmm0
  shufps xmm1, xmm1, $11
  addps xmm0, xmm1
  rsqrtps xmm0, xmm0       //approximate reciprocal
  mulps xmm2, xmm0
  movups [Result], xmm2

end;

{%endregion%}

{.$ENDIF}
{.$ELSE}
{%region%====[ Pure Pascal Vectors Functions ]==================================}

procedure TGLZVector4f.Create(Const aX,aY,aZ,aW : Single);
begin
   Self.X := AX;
   Self.Y := AY;
   Self.Z := AZ;
   Self.W := AW;
end;

function TGLZVector4f.ToString : String;
begin
   Result := '(X: '+FloattoStrF(Self.X,fffixed,5,5)+
            ' ,Y: '+FloattoStrF(Self.Y,fffixed,5,5)+
            ' ,Z: '+FloattoStrF(Self.Z,fffixed,5,5)+
            ' ,W: '+FloattoStrF(Self.W,fffixed,5,5)+')';
End;

class operator TGLZVector4f.+(constref A, B: TGLZVector4f): TGLZVector4f;
begin
  Result.X := A.X + B.X;
  Result.Y := A.Y + B.Y;
  Result.Z := A.Z + B.Z;
  Result.W := A.W + B.W;
end;

class operator TGLZVector4f.-(constref A, B: TGLZVector4f): TGLZVector4f;
begin
  Result.X := A.X - B.X;
  Result.Y := A.Y - B.Y;
  Result.Z := A.Z - B.Z;
  Result.W := A.W - B.W;
end;

class operator TGLZVector4f.*(constref A, B: TGLZVector4f): TGLZVector4f;
begin
  Result.X := A.X * B.X;
  Result.Y := A.Y * B.Y;
  Result.Z := A.Z * B.Z;
  Result.W := A.W * B.W;
end;

class operator TGLZVector4f./(constref A, B: TGLZVector4f): TGLZVector4f;
begin
  Result.X := A.X / B.X;
  Result.Y := A.Y / B.Y;
  Result.Z := A.Z / B.Z;
  Result.W := A.W / B.W;
end;

class operator TGLZVector4f.-(constref A : TGLZVector4f): TGLZVector4f;
begin
  Result.X := -A.X;
  Result.Y := -A.Y;
  Result.Z := -A.Z;
  Result.W := -A.W;
end;

class operator TGLZVector4f.+(constref A: TGLZVector4f; constref B:Single): TGLZVector4f; overload;
begin
  Result.X := A.X + B;
  Result.Y := A.Y + B;
  Result.Z := A.Z + B;
  Result.W := A.W + B;
end;

class operator TGLZVector4f.-(constref A: TGLZVector4f; constref B:Single): TGLZVector4f; overload;
begin
  Result.X := A.X - B;
  Result.Y := A.Y - B;
  Result.Z := A.Z - B;
  Result.W := A.W - B;
end;

class operator TGLZVector4f.*(constref A: TGLZVector4f; constref B:Single): TGLZVector4f; overload;
begin
  Result.X := A.X * B;
  Result.Y := A.Y * B;
  Result.Z := A.Z * B;
  Result.W := A.W * B;
end;

class operator TGLZVector4f./(constref A: TGLZVector4f; constref B:Single): TGLZVector4f; overload;
begin
  Result.X := A.X / B;
  Result.Y := A.Y / B;
  Result.Z := A.Z / B;
  Result.W := A.W / B;
end;

function TGLZVector4f.DivideBy2 : TGLZVector4f;
begin
  Result.x := Self.X*0.5;
  Result.y := Self.Y*0.5;
  Result.z := Self.Z*0.5;
  Result.w := Self.W*0.5;
end;

function TGLZVector4f.Distance(constref A: TGLZVector4f):Single;
begin
  result := sqrt(sqr(Self.X-A.X)+ sqr(Self.Y-A.Y) + sqr(Self.Z-A.Z));
end;

function TGLZVector4f.Length : Single;
//var
//X1 : Single;
begin
  Result := Sqrt((Self.X * Self.X) +(Self.Y * Self.Y) +(Self.Z * Self.Z));
  { Ensure result is non-zero to avoid divide-by-zero errors }
  //if X1 = 0 then X1 := 0.0000000001;
  //Result := X1;
end;

function TGLZVector4f.DotProduct(constref A: TGLZVector4f) : Single;
var
 V2:TGLZVector4f;
begin
  V2.X := Self.X*A.X;
  V2.Y := Self.Y*A.Y;
  V2.Z := Self.Z*A.Z;
 // V2.W := Self.W*A.W;
  Result := V2.X + V2.Y + V2.Z; // + V2.W;
end;

function TGLZVector4f.CrossProduct(Constref A : TGLZVector4f):TGLZVector4f;
begin
  result.X:=Self.Y*A.Z-Self.Z*A.Y;
  result.Y:=Self.Z*A.X-Self.X*A.Z;
  result.Z:=Self.X*A.Y-Self.Y*A.X;
  result.W:=1;
end;

function TGLZVector4f.Norm : Single;
var v2:TGLZVector4f;
begin
  v2.x := Self.X*Self.X;
  v2.y := Self.Y*Self.Y;
  v2.z := Self.Z*Self.Z;
  v2.w := Self.W*Self.W;
  Result := v2.X + v2.Y + v2.Z + v2.w;
end;

function TGLZVector4f.Normalize : TGLZVector4f;
var
  invLen : Single;
  vn : single;
begin
  vn:=Self.Norm;
  if vn=0 then
  begin
    result := Self;
    result.W := 0; //cZero;
  end
  else
  begin
    invLen:=1/Sqrt(vn);
    result.X:=Self.X*invLen;
    result.Y:=Self.Y*invLen;
    result.Z:=Self.Z*invLen;
    result.W:=Self.W;
  end;
end;
{.$ENDIF}

{%endregion%}

End.

